#ifndef Chip_cc
#define Chip_cc

#include <mathconst/mathconst.h>

#include "Chip.h"

Chip::Chip( const string& id, const dvm3_Vector cpccpc[],
		  const dvm3_Vector cpcclipcpclip[], const dvm3_Vector lsilsi[],
		  const ChipSize& cs, const bool infplane, 
		  const size_t num ) throw ( DeticptException ) : 
  Plane<rl_BasicRay>( lsilsi[ 0 ], lsilsi[ 1 ], lsilsi[ 2 ] ) {

  try { 

    identification = id;
    chipSize = cs;
    inf_plane = infplane;
    
    for ( size_t ii = 0; ii < num; ii++ ) {
      cpc[ ii ] = cpccpc[ ii ];
      cpcclip[ ii ] = cpcclipcpclip[ ii ];
      lsi[ ii ] = lsilsi[ ii ];
    }
    
    setup_matrices( );

  } catch ( DeticptException& de ) {

    throw( de );

  }
}

Chip& Chip::operator = ( const Chip& rhs ) {

  // check for assignment against itself
  if ( this == &rhs )
    return *this;
  this->Plane<rl_BasicRay>::operator = ( rhs );

  identification = rhs.identification;
  chipSize = rhs.chipSize;
  inf_plane = rhs.inf_plane;

  for ( size_t ii = 0; ii < NUMPOINT4; ii++ ) {
    cpc[ ii ] = rhs.cpc[ ii ];
    cpcclip[ ii ] = rhs.cpcclip[ ii ];
    lsi[ ii ] = rhs.lsi[ ii ];
  }

  osacmnc_rotation    = rhs.osacmnc_rotation;
  mncfc_rotation      = rhs.mncfc_rotation;
  fcstf_rotation      = rhs.fcstf_rotation;
  stflsi_rotation     = rhs.stflsi_rotation;
  lsicpc_rotation     = rhs.lsicpc_rotation;
  
  osacmnc_translation = rhs.osacmnc_translation;
  mncfc_translation   = rhs.mncfc_translation;
  fcstf_translation   = rhs.fcstf_translation;
  stflsi_translation  = rhs.stflsi_translation;
  lsicpc_translation  = rhs.lsicpc_translation;

  cpclsi_rotmat       = rhs.cpclsi_rotmat;

  return *this;

}

/**
 *
 * Determine if the intersection point,
 * intersectpos, falls within the chip.
 *
 */
Plane<rl_BasicRay>::MissHit Chip::clip_chip( const dvm3_Vector& intersectdir,
					    const dvm3_Vector& intersectpos,
					    const DetXY& detXY,
					    const MirrorPars& mirrorPars,
					    const dvm3_Vector& aimpointOSAC,
					    DeticptResult& deticptResult )
  const {

  dvm3_Vector cpcdir;
  dvm3_Vector cpcpos(0,0,0);
  // Transform the directional vector of the photon to the cpc coordinate
  lsicpc_rotation.mvmult( cpcdir, intersectdir ); 
  cpcdir.unitize( );
  // Transform the position of the photon to the cpc coordinate.
  lsicpc_rotation.mvmult( cpcpos, intersectpos - lsi[LL] );

  double xxx = cpcpos[0];
  double yyy = cpcpos[1];

  deticptResult.chip_id   = chipSize.get_chip_id( );
  deticptResult.cpcdir.x  = cpcdir[0];
  deticptResult.cpcdir.y  = cpcdir[1];
  deticptResult.cpcdir.z  = cpcdir[2];
  // You get the non-pixelized cpc for free
  deticptResult.cpcpos_np.x = cpcpos[0];
  deticptResult.cpcpos_np.y = cpcpos[1];
  deticptResult.cpcpos_np.z = cpcpos[2];

  // Calculate non-pixelized and pixelized chip
  deticptResult.chip_np.x   = 0.5 +
    deticptResult.cpcpos_np.x / chipSize.get_pixel_size_mm( ChipSize::XXX );
  deticptResult.chip_np.y   = 0.5 +
    deticptResult.cpcpos_np.y / chipSize.get_pixel_size_mm( ChipSize::YYY );
  deticptResult.chip.x      = (int)
    (deticptResult.cpcpos_np.x / chipSize.get_pixel_size_mm(
      ChipSize::XXX )) + 1;
  deticptResult.chip.y      = (int)
    (deticptResult.cpcpos_np.y / chipSize.get_pixel_size_mm(
      ChipSize::YYY )) + 1;

  // Calculate pixelized cpc
  deticptResult.cpcpos.x = ( deticptResult.chip.x - 0.5 ) *
    chipSize.get_pixel_size_mm( ChipSize::XXX );
  deticptResult.cpcpos.y = ( deticptResult.chip.y - 0.5 ) *
    chipSize.get_pixel_size_mm( ChipSize::YYY );
  deticptResult.cpcpos.z = deticptResult.cpcpos_np.z;

  // Now tdet non-pixelized and pixelized
  // combine the following two calls into one.
  deticptResult.tdet_np.x =
    chipSize.tdeltx( deticptResult.chip_np.x, deticptResult.chip_np.y );
  deticptResult.tdet_np.y =
    chipSize.tdelty( deticptResult.chip_np.x, deticptResult.chip_np.y );
  deticptResult.tdet.x =
    (int)(chipSize.tdeltx( deticptResult.chip.x, deticptResult.chip.y ));
  deticptResult.tdet.y =
    (int)(chipSize.tdelty( deticptResult.chip.x, deticptResult.chip.y ));
  
  // Convert cpc to lsi for both non-pixelized and pixelized
  dvm3_Vector intersectpos_pix( intersectpos );
  dvm3_Vector cpcpos_pix( deticptResult.cpcpos.x, deticptResult.cpcpos.y, deticptResult.cpcpos.z );
  lsicpc_rotation.mtvmult( intersectpos_pix, cpcpos_pix );
  intersectpos_pix += lsi[LL];
  
  // Calculate det for both non-pixelized and pixelized
  det_xy( intersectpos, aimpointOSAC, detXY, mirrorPars.focal_length,
	  deticptResult.det_np.x, deticptResult.det_np.y );
  det_xy( intersectpos_pix, aimpointOSAC, detXY, mirrorPars.focal_length,
	  deticptResult.det.x, deticptResult.det.y );
  
  // Calculate osac for both non-pixelized and pixelized
  deticptResult.position_np.x =
    +1 * ( intersectpos[1] + detXY.aimoffset[1] )
    + aimpointOSAC[0];
  deticptResult.position_np.y = 
    -1 * ( intersectpos[2] + detXY.aimoffset[2] )
    + aimpointOSAC[1];
  deticptResult.position_np.z = 
    -1 * ( intersectpos[0] + detXY.aimoffset[0] )
    + aimpointOSAC[2];
  
  deticptResult.position.x =
    +1 * ( intersectpos_pix[1] + detXY.aimoffset[1] )
    + aimpointOSAC[0];
  deticptResult.position.y = 
    -1 * ( intersectpos_pix[2] + detXY.aimoffset[2] )
    + aimpointOSAC[1];
  deticptResult.position.z = 
    -1 * ( intersectpos_pix[0] + detXY.aimoffset[0] )
    + aimpointOSAC[2];
  
  
  if ( cpcclip[ LL ][0] <= xxx && xxx <= cpcclip[ UR ][0] &&
       cpcclip[ LL ][1] <= yyy && yyy <= cpcclip[ UR ][1] )
    
    return Plane<rl_BasicRay>::HIT;
  
  else
    
    return Plane<rl_BasicRay>::MISS;

}

void Chip::check_cpc_to_lsi_transformation( ostream& os ) const {

  for ( size_t ii = 0; ii < NUMPOINT4; ii++ ) {

    dvm3_Vector result, xyzdiff;
    lsicpc_rotation.mtvmult( result, cpc[ ii ] );
    result += lsi[LL];
    xyzdiff = lsi[ ii ] - result;

    os << "The transformed - actual lsi position is : ";
    xyzdiff.print_on( os );
    os << '\n';
    
  }

}

void Chip::check_lsi_to_cpc_transformation( ostream& os ) const {

  for ( size_t ii = 0; ii < NUMPOINT4; ii++ ) {

    dvm3_Vector result, xyzdiff;

    lsicpc_rotation.mvmult( result, lsi[ ii ] - lsi[LL] );

    xyzdiff = cpc[ ii ] - result;

    os << "The transformed - actual cpc position is : ";
    xyzdiff.print_on( os );
    os << '\n';
    
  }

}

void Chip::det_xy( 
  const dvm3_Vector& lsiintersectpos, 
  const dvm3_Vector& aimpointOSAC,
  const DetXY& detXY,
  const double& focal_length, 
  double& detx, 
  double& dety ) 
  const {

  double fcx = lsiintersectpos[0] +                        detXY.aimoffset[0];
  double fcy = lsiintersectpos[1] +      aimpointOSAC[0] + detXY.aimoffset[1];
  double fcz = lsiintersectpos[2] + -1.0*aimpointOSAC[1] + detXY.aimoffset[2];

  double mnx = fcx - focal_length;
  double mny = fcy;
  double mnz = fcz;

  /* 
     "Coordinate Systems for Analysis of On-Orbit Chandra Data Paper
     1:  Imaging" states that the DET coord system is calculated as
     follows: 
       detx = detXY.X0 - detXY.delta * mny / mnx;
       dety = detXY.Y0 + detXY.delta * mnz / mnx;

     Looking at the units involved...
       pixels = pixels - ( radians / pixel ) * radians

     This, I believe indicates that there was a typo and the equations
     should be as below.
  */
  detx = detXY.X0 - 1 / detXY.delta * mny / mnx;
  dety = detXY.Y0 + 1 / detXY.delta * mnz / mnx;

}

void Chip::print( ostream& os, const char* str ) const {

  if ( str )
    os << str << endl;

  // print the normal, kkk etc...
  Plane<rl_BasicRay>::print( os );
  os << ", " << identification.c_str( ) << ", ";

  const char* tmp[] = { "LL", "LR", "UR", "UL" };

  for ( size_t ii = 0; ii < NUMPOINT4; ii++ ) {
    os << "cpc" << tmp[ ii ] << " = ";
    cpc[ ii ].print_on( os );
    os << ' ';
  }

  for ( size_t ii = 0; ii < NUMPOINT4; ii++ ) {
    os << "lsi" << tmp[ ii ] << " = ";
    lsi[ ii ].print_on( os );
    if ( ii != NUMPOINT4 - 1 )
      os << ' ';
  }

}

Plane<rl_BasicRay>::MissHit Chip::project_to_chip( 
  const dvm3_Vector& dir,
  const dvm3_Vector& pos,
  const DetXY& detXY,
  const MirrorPars& mirrorPars,
  const dvm3_Vector& aimpointOSAC,
  const dvm3_Vector& nodeOSAC,
  DeticptResult& deticptResult,
  double tol ) const {
  
  dvm3_Vector mncdir, fcdir, stfdir, lsidir, cpcdir;
  dvm3_Vector mncpos, fcpos, stfpos, lsipos, cpcpos;

  osacmnc_rotation.mvmult( mncdir, dir );
  osacmnc_rotation.mvmult( mncpos, pos - aimpointOSAC );

  mncfc_rotation.mvmult( fcdir, mncdir );
  mncfc_rotation.mvmult( fcpos, mncpos );

  fcstf_rotation.mvmult( stfdir, fcdir );
  fcstf_rotation.mvmult( stfpos, fcpos );

  stflsi_rotation.mvmult( lsidir, stfdir );
  stflsi_rotation.mvmult( lsipos, stfpos );
  lsipos -= detXY.aimoffset;

  // Transform the directional vector of the photon to the cpc coordinate
  lsicpc_rotation.mvmult( cpcdir, lsidir );
  cpcdir.unitize( );
  // Transform the position of the photon to the cpc coordinate.
  lsicpc_rotation.mvmult( cpcpos, lsipos - lsi[LL] );

  dvm3_Vector intersectpos;
  rl_BasicRay lsiray( lsipos, lsidir, 0.0, 0 );

  Plane<rl_BasicRay>::MissHit hitormiss_plane =
    intersect( lsiray, intersectpos, tol );

  // It may have hit the plane but that does not
  // mean that it is within the chip
  //  if ( ( ! inf_plane ) && Plane<RayType>::HIT == hitormiss_plane )
  //    return clip_chip( lsidir, intersectpos, detXY, mirrorPars,
  //		      aimpointOSAC, deticptResult ); 

  // It may have hit the plane but that does not
  // mean that it is within the chip
  Plane<rl_BasicRay>::MissHit hitormiss_chip = 
    clip_chip( lsidir, intersectpos, detXY, mirrorPars, aimpointOSAC,
	       deticptResult );

  return ( inf_plane ? hitormiss_plane : hitormiss_chip );

}

void Chip::setup_matrices( ) throw ( DeticptException ) {

  // OSAC -> MNC
  osacmnc_rotation.init( -1.0*M_PI/2.0, M_PI/2.0, 0.0, dvm3_RotMat::EY_convention );
  /* osacmnc_translation.init( node and aimpoint ); */

  // MNC  -> FC
  mncfc_rotation.init( 0.0, 0.0, 0.0, dvm3_RotMat::EY_convention );
  /* mncfc_translation.init( focal length ); */

  // FC   -> STF
  fcstf_rotation.init( 0.0, 0.0, 0.0, dvm3_RotMat::EY_convention );
  fcstf_translation.init( 0.0, 0.0, 0.0 );

  // STF  -> LSI
  stflsi_rotation.init( 0.0, 0.0, 0.0, dvm3_RotMat::EY_convention );
  /* stflsi_translation.init( aimoffset ); */

  // basis vectors of CPC Coordinate system, expressed in LSI coords
  dvm3_Vector exlsi, eylsi, ezlsi;
  exlsi = lsi[ LR ] - lsi[ LL ];
  eylsi = lsi[ UL ] - lsi[ LL ];
  exlsi.unitize( );
  eylsi.unitize( );
  ezlsi.cross( exlsi, eylsi );

  cpclsi_rotmat.init_by_col( exlsi, eylsi, ezlsi );

  lsicpc_rotation.init_by_row( exlsi, eylsi, ezlsi );
  /* lsicpc_translation.init( lsi[ LL ] ); */

}

void Chip::rotation_times_itsinverse( ostream& os ) const {

  dvm3_Matrix rotmat = lsicpc_rotation;
  dvm3_Matrix transpose;
  for ( size_t ii = 0; ii < 3; ii++ )
    for ( size_t jj = 0; jj < 3; jj++ ) 
      transpose[jj][ii] = rotmat[ii][jj];

  os << "The rotation matrix :"         << endl
     << transpose                          << endl;

  os << "The inverse rotation matrix :" << endl
     << rotmat                       << endl;
  dvm3_Matrix result;
  mmult(result,transpose,rotmat);
  os << "The product of the matrices :" << endl
     << result              << endl;

}

#endif
