#ifndef Chip_H
#define Chip_H

#include <deque>
#include <functional>
#include <iomanip>
#include <iostream>

#include <dvm3/dvm3.h>
#include <rl_basicray/rl_BasicRay.h>
#include <suplib/str.h>
#include <Plane/Plane.h>

#include "ChipSize.h"
#include "DetXY.h"
#include "DeticptResult.h"
#include "DeticptException.h"
#include "MirrorPars.h"

using namespace std;

/**
   The class Chip is responsible for transforming the photon to the 
   chip physical coordinate (cpc) system and determines if the photon
   falls within the chip or not.  If the photon falls on the chip then
   the class also writes out some info (intersection point and directional
   vector in various coordinates) to the binary pipe.
   Perhaps, the writing of the info should be moved up to the Detector class?
 */
class Chip : public Plane<rl_BasicRay> {

  friend ostream& operator << ( ostream& os, Chip& a ) {
    a.print( os );
    return os;
  }

  friend ostream& operator << ( ostream& os, Chip* a ) {
    a->print( os );
    return os;
  }

public:

  enum NumPoint { NUMPOINT4 = 4 };

  enum Corner { LL, LR, UR, UL };

  Chip( ) : Plane<rl_BasicRay>( ) { }

  Chip( const Chip& c ) { *this = c; }

  Chip& operator = ( const Chip& rhs );

  Chip( const string& id, const dvm3_Vector c[], const dvm3_Vector clip[], const dvm3_Vector l[],
	const ChipSize& cs, const bool infplane,
	const size_t num=NUMPOINT4 ) throw ( DeticptException );

  /// Determine if the photon falls within the chip or not?
  Plane<rl_BasicRay>::MissHit project_to_chip( const dvm3_Vector& dir, const dvm3_Vector& pos,
					       const DetXY& detXY,
					       const MirrorPars& mirrorPars,
					       const dvm3_Vector& aimpointOSAC,
					       const dvm3_Vector& nodeOSAC,
					       DeticptResult& deticptResult,
					       double tol=100.0*DBL_EPSILON ) const;

  dvm3_Vector get_cpc( int i ) const { return cpc[ i % NUMPOINT4 ]; }

  dvm3_Vector get_lsi( int i ) const { return lsi[ i % NUMPOINT4 ]; }

  string get_id( ) const { return identification; }

  virtual void print( ostream& os=cerr, const char* str=NULL ) const ;

  void check_cpc_to_lsi_transformation( ostream& os ) const;

  void check_lsi_to_cpc_transformation( ostream& os ) const;

  void rotation_times_itsinverse( ostream& os ) const;

protected:

  /// The name of the chip
  string identification;

  /// The chip physical coordinates
  dvm3_Vector cpc[ 4 ];

  /// The chip cpc coordinate to clip
  dvm3_Vector cpcclip[ 4 ];

  /// The local scientific instrument coordinates
  dvm3_Vector lsi[ 4 ];

  ChipSize chipSize;

  dvm3_RotMat osacmnc_rotation;
  dvm3_RotMat mncfc_rotation;
  dvm3_RotMat fcstf_rotation;
  dvm3_RotMat stflsi_rotation;
  dvm3_RotMat lsicpc_rotation;

  dvm3_Vector osacmnc_translation;
  dvm3_Vector mncfc_translation;
  dvm3_Vector fcstf_translation;
  dvm3_Vector stflsi_translation;
  dvm3_Vector lsicpc_translation;

  dvm3_Matrix cpclsi_rotmat;
  
  bool inf_plane;
  
private:

  Plane<rl_BasicRay>::MissHit clip_chip( const dvm3_Vector& intersectdir,
				  const dvm3_Vector& intersectpos,
				  const DetXY& detXY,
				  const MirrorPars& mirrorPars,
				  const dvm3_Vector& aimpointOSAC,
				  DeticptResult& deticptResult ) const;

  /// Transform from cpc to lsi coordinate system.
  void cpc_to_lsi( const dvm3_Vector& from, dvm3_Vector& to ) const;

  /// Transform from lsi to cpc coordinate system.
  void lsi_to_cpc( const dvm3_Vector& from, dvm3_Vector& to ) const;

  /// Transform direction vector from lsi to cpc coordinate system.
  void lsi_to_cpc_dir( const dvm3_Vector& from, dvm3_Vector& to ) const;

  /// Set up the RCPLSI and RCPLSI_inverse matrix.
  void setup_matrices( ) throw( DeticptException );

  // Calculate the detector (focal plane pixel) coordinates.
  void det_xy(
    const dvm3_Vector& lsiintersectpos,
    const dvm3_Vector& aimpointOSAC,
    const DetXY& detXY, 
    const double& focal_length,
    double& detx,
    double& dety ) const;

};

#endif
