#ifndef Detector_CC
#define Detector_CC

#include <iostream>

#include <suplibxx/str.h>
#include <rdbxx/RDB.h>
#include <rdbxx/simpleRDBTable.h>

#include "Detector.h"

Detector::Detector( ParFile& parfile, const char* rdbheader[],
			      const char* corners_id[],
			      const ChipSizeMap& chipSizeMap )
  throw ( ParFileException, DeticptException, Exception ) {

  try {

    init_Detector( parfile, rdbheader, corners_id, chipSizeMap );

  } catch( DeticptException& de ) {

    throw;

  } catch( ParFileException& pfe ) {

    throw;

  } catch( Exception& e ) {

    throw;

  }

}

/**
 * Make sure that the corner id is one of the entries of corners_id.
 */
void Detector::check_corner_id( const string& str,
					  const char* corners_id[] ) const
  throw( DeticptException ) {
  
  for ( size_t ii = 0; NULL != corners_id[ ii ]; ii++ )
    if ( 0 == str.compare( corners_id[ ii ] ) )
      return;

  const char* format=
    "Detector<D>::check_corner_id( ) : `%s' is not a valid option "
    "for a corner id\n";

  throw DeticptException( format, str.c_str( ) );

}

/**
 * Make sure that the transformation from the fiducial cpc and lsi are legit.
 */
void Detector::diagnostics_only( ostream& os ) const {

  for ( vector<Chip>::const_iterator iter =
	  the_chips.begin( ); iter != the_chips.end( ); ++iter ) {

    os << "For chip : " << iter->get_id( ) << '\n';
    iter->rotation_times_itsinverse( os );
    iter->check_cpc_to_lsi_transformation( os );
    iter->check_lsi_to_cpc_transformation( os );
    os << endl;

  }

}

void Detector::init_Detector( ParFile& parfile,
			      const char* rdbheader[],
			      const char* corners_id[],
			      const ChipSizeMap& chipSizeMap )
  throw( ParFileException, DeticptException, Exception ) {

  try {

    filename    = parfile.pgetstring( "chipposdb" );
    // cerr << "filename = `" << filename << "'\n";
    quick_check = parfile.pgetb( "quickcheck" );
    inf_plane   = parfile.pgetb( "infiniteplane" );

    myMap chips;
    mySet unique_chip_id;
    read_database( parfile, rdbheader, corners_id, chips, unique_chip_id );
    make_Detector( chips, corners_id, unique_chip_id, chipSizeMap );

  } catch( ParFileException& pfe ) {

    throw;

  } catch( DeticptException& de ) {

    throw;

  } catch( Exception& e ) {

    throw;

  }

}

/**
 *
 * For each chip id extract the corners in the order of corners_id.
 * Create a chip in with the corners as detailed in the corners_id
 * string array.  Insert the chip in the vector the_chips.
 *
 */
void Detector::make_Detector( myMap& chips, const char* corners_id[],
					const mySet& unique_chip_id,
					const ChipSizeMap& chipSizeMap )
  throw ( DeticptException ) {

  try { 
    for ( mySet::const_iterator iter = unique_chip_id.begin( ); 
	  iter != unique_chip_id.end( ); ++iter ) {
      
      dvm3_Vector cpc2[ Chip::NUMPOINT4 ];
      dvm3_Vector lsi2[ Chip::NUMPOINT4 ];
      dvm3_Vector cpcclip2[ Chip::NUMPOINT4 ];
      
      string id( *iter );
      
      // Loop over the 4 corners of a chip.
      for ( size_t ii = 0; NULL != corners_id[ ii ]; ii++ ) {
	
	string id_corners( *iter );
	id_corners += corners_id[ ii ];
	
	map<string, dbChipRow>::const_iterator val_ptr; 
	val_ptr = chips.find( id_corners );
	
	if ( val_ptr != chips.end( ) ) {
	  
	  cpc2[ ii ]     = val_ptr->second.get_cpc( );
	  cpcclip2[ ii ] = val_ptr->second.get_cpc_clip( );
	  lsi2[ ii ]     = val_ptr->second.get_lsi( );
	  string key( val_ptr->first );
	  
	} else {
	  
	  const char* format =
	    "void Detector<D>::init_Detector( myMap&, const char*, "
	    "const mySet&, const ChipSizeMap& ) : Unable to find `%s'\n";
	  throw DeticptException( format, id_corners.c_str( ) );
	  
	}
	
      }
      
      ChipSizeMapIterator chipsize_ptr = chipSizeMap.find( id );
      if ( chipsize_ptr == chipSizeMap.end( ) ) {
	
	const char* format =
	  "void Detector<D>::init_Detector( myMap&, const char*, "
	  "const mySet&, const ChipSizeMap& ) : Unable to find `%s' in the\n";
	throw DeticptException( format, id.c_str( ) );
	
      }
      
      Chip chip( id, cpc2, cpcclip2, lsi2,
			    chipsize_ptr->second,
			    inf_plane,
				       Chip::NUMPOINT4 );
      
      the_chips.push_back( chip );
      
    }
  } catch ( DeticptException& de ) {

    throw( de );

  }
}

/**
 * Print the content of the object.
 */
void Detector::print( ostream& os ) const {

  for ( vector<Chip>::const_iterator iter = the_chips.begin( );
	iter != the_chips.end( ); ++iter ) {

    iter->print( os );

    vector<Chip>::const_iterator current = iter;
    if ( ++current != the_chips.end( ) )
      os << '\n';

  }

}

/**
 *  Loop thru the vector the_chips to find
 * if the photon lands on the chip or not.
 */
int Detector::project_to_detector( 
  const dvm3_Vector& dir,
  const dvm3_Vector& pos,
  const MirrorPars& mirrorPars,
  const dvm3_Vector& aimpointOSAC,
  const dvm3_Vector& nodeOSAC,
  DeticptResult& deticptResult,
  double tol ) const {
  
  DeticptResult minResult;
  Plane<rl_BasicRay>::MissHit hitormiss_any = Plane<rl_BasicRay>::MISS;

  /// Loop through the vector of chips to see where the photon lands.
  for ( vector<Chip>::const_iterator iter = the_chips.begin( );
	iter != the_chips.end( ); ++iter ) {

    // presumably it can land only one chip so return if it landed on the chip
    Plane<rl_BasicRay>::MissHit hitormiss = 
      iter->project_to_chip( dir, pos, detXY, mirrorPars,
			     aimpointOSAC, nodeOSAC, deticptResult, tol );

    if ( Plane<rl_BasicRay>::HIT == hitormiss )
    {
      if ( quick_check )
	return EXIT_SUCCESS;

      else {
	if ( Plane<rl_BasicRay>::MISS == hitormiss_any ) {
	  hitormiss_any = hitormiss;
	  minResult     = deticptResult;

	} else {
	  if ( deticptResult.position_np.z < minResult.position_np.z )
	    minResult = deticptResult;

	}
      }
    }
  }

  if ( Plane<rl_BasicRay>::HIT == hitormiss_any ) {
    deticptResult = minResult;
    return EXIT_SUCCESS;

  } else {
    return EXIT_FAILURE;

  }
}

/**
 *  Read the rdb file to get the positions of the chips.
 *  Note that the chips could be put in any order, but the column
 *  corner must be one of the following options: LL, LR, UR, UL.
 */
void Detector::read_database( ParFile& parfile, const char* rdbheader[],
					const char* corners_id[], 
					myMap& chips, mySet& unique_chip_id )
  throw ( DeticptException, ParFileException, Exception ) {

  try {

    simpleRDBTable< dbChipRow > table( filename, ios::in,
						  rdbheader );

    detXY.delta = arcseconds_to_radians( suplib::str2d( table.getComment( "delta" ).getValue().c_str( ) ) );
    detXY.X0 = suplib::str2d( table.getComment( "X0" ).getValue().c_str( ) );
    detXY.Y0 = suplib::str2d( table.getComment( "Y0" ).getValue().c_str( ) );
    detXY.aimoffset[0] = suplib::str2d( table.getComment( "xaimoffset" ).getValue().c_str( ) );
    detXY.aimoffset[1] = suplib::str2d( table.getComment( "yaimoffset" ).getValue().c_str( ) );
    detXY.aimoffset[2] = suplib::str2d( table.getComment( "zaimoffset" ).getValue().c_str( ) );

    one_of_three_ways( parfile.pgetstring( "xaimoffset" ), detXY.aimoffset[0] );
    one_of_three_ways( parfile.pgetstring( "yaimoffset" ), detXY.aimoffset[1] );
    one_of_three_ways( parfile.pgetstring( "zaimoffset" ), detXY.aimoffset[2] );

    dbChipRow* ptr = NULL;
    while( NULL != (ptr=table.readRow( )) ) {

      string chip_id( ptr->get_id( ) );
      string corner_id( ptr->get_corner( ) );

      check_corner_id( corner_id, corners_id );

      string chip_corner_id = chip_id + corner_id;
      unique_chip_id.insert( chip_id );

      chips[ chip_corner_id ] = *ptr;

      delete ptr;
      ptr = NULL;

    }

    /*
    for ( DetectorSetIterator iter = unique_chip_id.begin( ); 
	  iter != unique_chip_id.end( ); ++iter )
      cout << *iter << endl;
    */

  } catch( RDBErr& re ) {

    cerr << "Detector<T>::read_database( const char*, const char**,"
      "myMap&, mySet& ) : ";
    cerr << re.what( ) << '\n';
    exit( EXIT_FAILURE );

  } catch( DeticptException& de ) {

    throw;

  } catch( Exception& e ) {

    throw;

  }

}

#endif // #ifndef Detector_CC
